@@ -5,14 +5,16 @@ use crate::compile::warn_missing_py_init;
 use crate::module_writer::{
     add_data, write_bin, write_bindings_module, write_cffi_module, write_python_part, WheelWriter,
 };
+use crate::project_layout::ProjectLayout;
 use crate::source_distribution::source_distribution;
-use crate::{compile, Metadata21, ModuleWriter, PyProjectToml, PythonInterpreter, Target};
+use crate::{
+    compile, BuildArtifact, Metadata21, ModuleWriter, PyProjectToml, PythonInterpreter, Target,
+};
 use anyhow::{anyhow, bail, Context, Result};
 use cargo_metadata::Metadata;
 use fs_err as fs;
 use lddtree::Library;
 use sha2::{Digest, Sha256};
-use std::borrow::Cow;
 use std::collections::{HashMap, HashSet};
 use std::fmt::{Display, Formatter};
 use std::io;
@@ -72,93 +74,6 @@ impl Display for BridgeModel {
     }
 }
 
-/// Whether this project is pure rust or rust mixed with python and whether it has wheel data
-#[derive(Clone, Debug, PartialEq, Eq)]
-pub struct ProjectLayout {
-    /// Contains the canonicalized (i.e. absolute) path to the python part of the project
-    /// If none, we have a rust crate compiled into a shared library with only some glue python for cffi
-    /// If some, we have a python package that is extended by a native rust module.
-    pub python_module: Option<PathBuf>,
-    /// Contains the canonicalized (i.e. absolute) path to the rust part of the project
-    pub rust_module: PathBuf,
-    /// rust extension name
-    pub extension_name: String,
-    /// The location of the wheel data, if any
-    pub data: Option<PathBuf>,
-}
-
-impl ProjectLayout {
-    /// Checks whether a python module exists besides Cargo.toml with the right name
-    pub fn determine(
-        project_root: impl AsRef<Path>,
-        module_name: &str,
-        py_src: Option<impl AsRef<Path>>,
-        data: Option<impl AsRef<Path>>,
-    ) -> Result<ProjectLayout> {
-        // A dot in the module name means the extension module goes into the module folder specified by the path
-        let parts: Vec<&str> = module_name.split('.').collect();
-        let project_root = project_root.as_ref();
-        let python_root = py_src.map_or(Cow::Borrowed(project_root), |py_src| {
-            Cow::Owned(project_root.join(py_src))
-        });
-        let (python_module, rust_module, extension_name) = if parts.len() > 1 {
-            let mut rust_module = python_root.to_path_buf();
-            rust_module.extend(&parts[0..parts.len() - 1]);
-            (
-                python_root.join(parts[0]),
-                rust_module,
-                parts[parts.len() - 1].to_string(),
-            )
-        } else {
-            (
-                python_root.join(module_name),
-                python_root.join(module_name),
-                module_name.to_string(),
-            )
-        };
-
-        let data = if let Some(data) = data {
-            let data = if data.as_ref().is_absolute() {
-                data.as_ref().to_path_buf()
-            } else {
-                project_root.join(data)
-            };
-            if !data.is_dir() {
-                bail!("No such data directory {}", data.display());
-            }
-            Some(data)
-        } else if project_root.join(format!("{}.data", module_name)).is_dir() {
-            Some(project_root.join(format!("{}.data", module_name)))
-        } else {
-            None
-        };
-
-        if python_module.is_dir() {
-            if !python_module.join("__init__.py").is_file()
-                && !python_module.join("__init__.pyi").is_file()
-            {
-                bail!("Found a directory with the module name ({}) next to Cargo.toml, which indicates a mixed python/rust project, but the directory didn't contain an __init__.py file.", module_name)
-            }
-
-            println!("🍹 Building a mixed python/rust project");
-
-            Ok(ProjectLayout {
-                python_module: Some(python_module),
-                rust_module,
-                extension_name,
-                data,
-            })
-        } else {
-            Ok(ProjectLayout {
-                python_module: None,
-                rust_module: project_root.to_path_buf(),
-                extension_name,
-                data,
-            })
-        }
-    }
-}
-
 /// Contains all the metadata required to build the crate
 #[derive(Clone)]
 pub struct BuildContext {
@@ -168,6 +83,10 @@ pub struct BuildContext {
     pub bridge: BridgeModel,
     /// Whether this project is pure rust or rust mixed with python
     pub project_layout: ProjectLayout,
+    /// The path to pyproject.toml. Required for the source distribution
+    pub pyproject_toml_path: PathBuf,
+    /// Parsed pyproject.toml if any
+    pub pyproject_toml: Option<PyProjectToml>,
     /// Python Package Metadata 2.1
     pub metadata21: Metadata21,
     /// The name of the crate
@@ -259,32 +178,23 @@ impl BuildContext {
         fs::create_dir_all(&self.out)
             .context("Failed to create the target directory for the source distribution")?;
 
-        let include_cargo_lock = self.cargo_options.locked || self.cargo_options.frozen;
-        match PyProjectToml::new(self.manifest_path.parent().unwrap()) {
-            Ok(pyproject) => {
-                let sdist_path = source_distribution(
-                    &self.out,
-                    &self.metadata21,
-                    &self.manifest_path,
-                    &self.cargo_metadata,
-                    pyproject.sdist_include(),
-                    include_cargo_lock,
-                    self.project_layout.data.as_deref(),
-                )
-                .context("Failed to build source distribution")?;
+        match self.pyproject_toml.as_ref() {
+            Some(pyproject) => {
+                let sdist_path = source_distribution(self, pyproject)
+                    .context("Failed to build source distribution")?;
                 Ok(Some((sdist_path, "source".to_string())))
             }
-            Err(_) => Ok(None),
+            None => Ok(None),
         }
     }
 
     fn auditwheel(
         &self,
-        artifact: &Path,
+        artifact: &BuildArtifact,
         platform_tag: &[PlatformTag],
         python_interpreter: Option<&PythonInterpreter>,
     ) -> Result<(Policy, Vec<Library>)> {
-        if self.skip_auditwheel || self.editable {
+        if self.skip_auditwheel {
             return Ok((Policy::default(), Vec::new()));
         }
 
@@ -322,12 +232,42 @@ impl BuildContext {
         get_policy_and_libs(artifact, tag, &self.target)
     }
 
+    /// Add library search paths in Cargo target directory rpath when building in editable mode
+    fn add_rpath(&self, artifacts: &[&BuildArtifact]) -> Result<()> {
+        if self.editable && self.target.is_linux() {
+            for artifact in artifacts {
+                if artifact.linked_paths.is_empty() {
+                    continue;
+                }
+                let old_rpaths = patchelf::get_rpath(&artifact.path)?;
+                let mut new_rpaths = old_rpaths.clone();
+                for path in &artifact.linked_paths {
+                    if !old_rpaths.contains(path) {
+                        new_rpaths.push(path.to_string());
+                    }
+                }
+                let new_rpath = new_rpaths.join(":");
+                if let Err(err) = patchelf::set_rpath(&artifact.path, &new_rpath) {
+                    println!(
+                        "⚠️ Warning: Failed to set rpath for {}: {}",
+                        artifact.path.display(),
+                        err
+                    );
+                }
+            }
+        }
+        Ok(())
+    }
+
     fn add_external_libs(
         &self,
         writer: &mut WheelWriter,
-        artifacts: &[PathBuf],
+        artifacts: &[&BuildArtifact],
         ext_libs: &[Vec<Library>],
     ) -> Result<()> {
+        if self.editable {
+            return self.add_rpath(artifacts);
+        }
         if ext_libs.iter().all(|libs| libs.is_empty()) {
             return Ok(());
         }
@@ -391,7 +331,7 @@ impl BuildContext {
                 })
                 .collect::<Vec<_>>();
             if !replacements.is_empty() {
-                patchelf::replace_needed(artifact, &replacements[..])?;
+                patchelf::replace_needed(&artifact.path, &replacements[..])?;
             }
         }
 
@@ -423,14 +363,13 @@ impl BuildContext {
         // Currently artifact .so file always resides at ${module_name}/${module_name}.so
         let artifact_dir = Path::new(&self.module_name);
         for artifact in artifacts {
-            let old_rpaths = patchelf::get_rpath(artifact)?;
+            let mut new_rpaths = patchelf::get_rpath(&artifact.path)?;
             // TODO: clean existing rpath entries if it's not pointed to a location within the wheel
             // See https://github.com/pypa/auditwheel/blob/353c24250d66951d5ac7e60b97471a6da76c123f/src/auditwheel/repair.py#L160
-            let mut new_rpaths: Vec<&str> = old_rpaths.split(':').collect();
             let new_rpath = Path::new("$ORIGIN").join(relpath(&libs_dir, artifact_dir));
-            new_rpaths.push(new_rpath.to_str().unwrap());
+            new_rpaths.push(new_rpath.to_str().unwrap().to_string());
             let new_rpath = new_rpaths.join(":");
-            patchelf::set_rpath(artifact, &new_rpath)?;
+            patchelf::set_rpath(&artifact.path, &new_rpath)?;
         }
         Ok(())
     }
@@ -444,7 +383,7 @@ impl BuildContext {
 
     fn write_binding_wheel_abi3(
         &self,
-        artifact: &Path,
+        artifact: BuildArtifact,
         platform_tags: &[PlatformTag],
         ext_libs: Vec<Library>,
         major: u8,
@@ -456,13 +395,13 @@ impl BuildContext {
         let tag = format!("cp{}{}-abi3-{}", major, min_minor, platform);
 
         let mut writer = WheelWriter::new(&tag, &self.out, &self.metadata21, &[tag.clone()])?;
-        self.add_external_libs(&mut writer, &[artifact.to_path_buf()], &[ext_libs])?;
+        self.add_external_libs(&mut writer, &[&artifact], &[ext_libs])?;
 
         write_bindings_module(
             &mut writer,
             &self.project_layout,
             &self.module_name,
-            artifact,
+            &artifact.path,
             None,
             &self.target,
             self.editable,
@@ -499,7 +438,7 @@ impl BuildContext {
             self.platform_tag.clone()
         };
         let (wheel_path, tag) = self.write_binding_wheel_abi3(
-            &artifact,
+            artifact,
             &platform_tags,
             external_libs,
             major,
@@ -520,20 +459,20 @@ impl BuildContext {
     fn write_binding_wheel(
         &self,
         python_interpreter: &PythonInterpreter,
-        artifact: &Path,
+        artifact: BuildArtifact,
         platform_tags: &[PlatformTag],
         ext_libs: Vec<Library>,
     ) -> Result<BuiltWheelMetadata> {
         let tag = python_interpreter.get_tag(&self.target, platform_tags, self.universal2)?;
 
         let mut writer = WheelWriter::new(&tag, &self.out, &self.metadata21, &[tag.clone()])?;
-        self.add_external_libs(&mut writer, &[artifact.to_path_buf()], &[ext_libs])?;
+        self.add_external_libs(&mut writer, &[&artifact], &[ext_libs])?;
 
         write_bindings_module(
             &mut writer,
             &self.project_layout,
             &self.module_name,
-            artifact,
+            &artifact.path,
             Some(python_interpreter),
             &self.target,
             self.editable,
@@ -575,7 +514,7 @@ impl BuildContext {
             };
             let (wheel_path, tag) = self.write_binding_wheel(
                 python_interpreter,
-                &artifact,
+                artifact,
                 &platform_tags,
                 external_libs,
             )?;
@@ -602,39 +541,40 @@ impl BuildContext {
         &self,
         python_interpreter: Option<&PythonInterpreter>,
         extension_name: Option<&str>,
-    ) -> Result<PathBuf> {
+    ) -> Result<BuildArtifact> {
         let artifacts = compile(self, python_interpreter, &self.bridge)
             .context("Failed to build a native library through cargo")?;
-        let artifacts = artifacts
-            .get(0)
-            .context("Failed to build a native library through cargo")?;
+        let error_msg = "Cargo didn't build a cdylib. Did you miss crate-type = [\"cdylib\"] \
+                 in the lib section of your Cargo.toml?";
+        let artifacts = artifacts.get(0).context(error_msg)?;
 
-        let artifact = artifacts.get("cdylib").cloned().ok_or_else(|| {
-            anyhow!(
-                "Cargo didn't build a cdylib. Did you miss crate-type = [\"cdylib\"] \
-                 in the lib section of your Cargo.toml?",
-            )
-        })?;
+        let mut artifact = artifacts
+            .get("cdylib")
+            .cloned()
+            .ok_or_else(|| anyhow!(error_msg,))?;
 
         if let Some(extension_name) = extension_name {
-            warn_missing_py_init(&artifact, extension_name)
-                .context("Failed to parse the native library")?;
+            // globin has an issue parsing MIPS64 ELF, see https://github.com/m4b/goblin/issues/274
+            // But don't fail the build just because we can't emit a warning
+            let _ = warn_missing_py_init(&artifact.path, extension_name);
         }
 
         if self.editable || self.skip_auditwheel {
             return Ok(artifact);
         }
         // auditwheel repair will edit the file, so we need to copy it to avoid errors in reruns
-        let maturin_build = artifact.parent().unwrap().join("maturin");
+        let artifact_path = &artifact.path;
+        let maturin_build = artifact_path.parent().unwrap().join("maturin");
         fs::create_dir_all(&maturin_build)?;
-        let new_artifact = maturin_build.join(artifact.file_name().unwrap());
-        fs::copy(&artifact, &new_artifact)?;
-        Ok(new_artifact)
+        let new_artifact_path = maturin_build.join(artifact_path.file_name().unwrap());
+        fs::copy(artifact_path, &new_artifact_path)?;
+        artifact.path = new_artifact_path;
+        Ok(artifact)
     }
 
     fn write_cffi_wheel(
         &self,
-        artifact: &Path,
+        artifact: BuildArtifact,
         platform_tags: &[PlatformTag],
         ext_libs: Vec<Library>,
     ) -> Result<BuiltWheelMetadata> {
@@ -643,7 +583,7 @@ impl BuildContext {
             .get_universal_tags(platform_tags, self.universal2)?;
 
         let mut writer = WheelWriter::new(&tag, &self.out, &self.metadata21, &tags)?;
-        self.add_external_libs(&mut writer, &[artifact.to_path_buf()], &[ext_libs])?;
+        self.add_external_libs(&mut writer, &[&artifact], &[ext_libs])?;
 
         write_cffi_module(
             &mut writer,
@@ -651,7 +591,7 @@ impl BuildContext {
             self.manifest_path.parent().unwrap(),
             &self.target_dir,
             &self.module_name,
-            artifact,
+            &artifact.path,
             &self.interpreter[0].executable,
             self.editable,
         )?;
@@ -672,7 +612,7 @@ impl BuildContext {
         } else {
             self.platform_tag.clone()
         };
-        let (wheel_path, tag) = self.write_cffi_wheel(&artifact, &platform_tags, external_libs)?;
+        let (wheel_path, tag) = self.write_cffi_wheel(artifact, &platform_tags, external_libs)?;
 
         // Warn if cffi isn't specified in the requirements
         if !self
@@ -696,7 +636,7 @@ impl BuildContext {
     fn write_bin_wheel(
         &self,
         python_interpreter: Option<&PythonInterpreter>,
-        artifacts: &[PathBuf],
+        artifacts: &[BuildArtifact],
         platform_tags: &[PlatformTag],
         ext_libs: &[Vec<Library>],
     ) -> Result<BuiltWheelMetadata> {
@@ -725,15 +665,18 @@ impl BuildContext {
             }
         }
 
+        let mut artifacts_ref = Vec::with_capacity(artifacts.len());
         for artifact in artifacts {
+            artifacts_ref.push(artifact);
             // I wouldn't know of any case where this would be the wrong (and neither do
             // I know a better alternative)
             let bin_name = artifact
+                .path
                 .file_name()
                 .expect("Couldn't get the filename from the binary produced by cargo");
-            write_bin(&mut writer, artifact, &self.metadata21, bin_name)?;
+            write_bin(&mut writer, &artifact.path, &self.metadata21, bin_name)?;
         }
-        self.add_external_libs(&mut writer, artifacts, ext_libs)?;
+        self.add_external_libs(&mut writer, &artifacts_ref, ext_libs)?;
 
         self.add_pth(&mut writer)?;
         add_data(&mut writer, self.project_layout.data.as_deref())?;
@@ -751,6 +694,9 @@ impl BuildContext {
         let mut wheels = Vec::new();
         let artifacts = compile(self, python_interpreter, &self.bridge)
             .context("Failed to build a native library through cargo")?;
+        if artifacts.is_empty() {
+            bail!("Cargo didn't build a binary")
+        }
 
         let mut policies = Vec::with_capacity(artifacts.len());
         let mut ext_libs = Vec::new();
