@@ -1,4 +1,5 @@
 use crate::build_context::BridgeModel;
+use crate::target::RUST_1_64_0;
 use crate::{BuildContext, PlatformTag, PythonInterpreter, Target};
 use anyhow::{anyhow, bail, Context, Result};
 use fat_macho::FatWriter;
@@ -14,6 +15,19 @@ use std::str;
 /// without `PYO3_NO_PYTHON` environment variable
 const PYO3_ABI3_NO_PYTHON_VERSION: (u64, u64, u64) = (0, 16, 4);
 
+/// crate types excluding `bin`, `cdylib` and `proc-macro`
+const LIB_CRATE_TYPES: [&str; 4] = ["lib", "dylib", "rlib", "staticlib"];
+
+/// A cargo build artifact
+#[derive(Debug, Clone)]
+pub struct BuildArtifact {
+    /// Path to the build artifact
+    pub path: PathBuf,
+    /// Array of paths to include in the library search path, as indicated by
+    /// the `cargo:rustc-link-search` instruction.
+    pub linked_paths: Vec<String>,
+}
+
 /// Builds the rust crate into a native module (i.e. an .so or .dll) for a
 /// specific python version. Returns a mapping from crate type (e.g. cdylib)
 /// to artifact location.
@@ -21,9 +35,9 @@ pub fn compile(
     context: &BuildContext,
     python_interpreter: Option<&PythonInterpreter>,
     bindings_crate: &BridgeModel,
-) -> Result<Vec<HashMap<String, PathBuf>>> {
+) -> Result<Vec<HashMap<String, BuildArtifact>>> {
     let root_pkg = context.cargo_metadata.root_package().unwrap();
-    let targets: Vec<_> = root_pkg
+    let mut targets: Vec<_> = root_pkg
         .targets
         .iter()
         .filter(|target| match bindings_crate {
@@ -31,6 +45,19 @@ pub fn compile(
             _ => target.kind.contains(&"cdylib".to_string()),
         })
         .collect();
+    if targets.is_empty() && !bindings_crate.is_bin() {
+        // No `crate-type = ["cdylib"]` in `Cargo.toml`
+        // Let's try compile one of the target with `--crate-type cdylib`
+        let lib_target = root_pkg.targets.iter().find(|target| {
+            target
+                .kind
+                .iter()
+                .any(|k| LIB_CRATE_TYPES.contains(&k.as_str()))
+        });
+        if let Some(target) = lib_target {
+            targets.push(target);
+        }
+    }
     if context.target.is_macos() && context.universal2 {
         compile_universal2(context, python_interpreter, bindings_crate, &targets)
     } else {
@@ -44,7 +71,7 @@ fn compile_universal2(
     python_interpreter: Option<&PythonInterpreter>,
     bindings_crate: &BridgeModel,
     targets: &[&cargo_metadata::Target],
-) -> Result<Vec<HashMap<String, PathBuf>>> {
+) -> Result<Vec<HashMap<String, BuildArtifact>>> {
     let build_type = if bindings_crate.is_bin() {
         "bin"
     } else {
@@ -91,12 +118,13 @@ fn compile_universal2(
         })?;
         // Create an universal dylib
         let output_path = aarch64_artifact
+            .path
             .display()
             .to_string()
             .replace("aarch64-apple-darwin/", "");
         let mut writer = FatWriter::new();
-        let aarch64_file = fs::read(aarch64_artifact)?;
-        let x86_64_file = fs::read(x86_64_artifact)?;
+        let aarch64_file = fs::read(&aarch64_artifact.path)?;
+        let x86_64_file = fs::read(&x86_64_artifact.path)?;
         writer
             .add(aarch64_file)
             .map_err(|e| anyhow!("Failed to add aarch64 cdylib: {:?}", e))?;
@@ -108,7 +136,11 @@ fn compile_universal2(
             .map_err(|e| anyhow!("Failed to create universal cdylib: {:?}", e))?;
 
         let mut result = HashMap::new();
-        result.insert(build_type.to_string(), PathBuf::from(output_path));
+        let universal_artifact = BuildArtifact {
+            path: PathBuf::from(output_path),
+            ..x86_64_artifact
+        };
+        result.insert(build_type.to_string(), universal_artifact);
         universal_artifacts.push(result);
     }
     Ok(universal_artifacts)
@@ -119,7 +151,7 @@ fn compile_targets(
     python_interpreter: Option<&PythonInterpreter>,
     bindings_crate: &BridgeModel,
     targets: &[&cargo_metadata::Target],
-) -> Result<Vec<HashMap<String, PathBuf>>> {
+) -> Result<Vec<HashMap<String, BuildArtifact>>> {
     let mut artifacts = Vec::with_capacity(targets.len());
     for target in targets {
         artifacts.push(compile_target(
@@ -137,7 +169,7 @@ fn compile_target(
     python_interpreter: Option<&PythonInterpreter>,
     bindings_crate: &BridgeModel,
     binding_target: &cargo_metadata::Target,
-) -> Result<HashMap<String, PathBuf>> {
+) -> Result<HashMap<String, BuildArtifact>> {
     let target = &context.target;
 
     let mut cargo_rustc: cargo_options::Rustc = context.cargo_options.clone().into();
@@ -148,6 +180,19 @@ fn compile_target(
         cargo_rustc.release = true;
     }
 
+    // Add `--crate-type cdylib` if available
+    if binding_target
+        .kind
+        .iter()
+        .any(|k| LIB_CRATE_TYPES.contains(&k.as_str()))
+    {
+        // `--crate-type` is stable since Rust 1.64.0
+        // See https://github.com/rust-lang/cargo/pull/10838
+        if target.rustc_version.semver >= RUST_1_64_0 {
+            cargo_rustc.crate_type = vec!["cdylib".to_string()];
+        }
+    }
+
     let mut rust_flags = env::var_os("RUSTFLAGS");
 
     // We need to pass --bin / --lib
@@ -197,13 +242,28 @@ fn compile_target(
         }
     } else if target.is_emscripten() {
         let flags = rust_flags.get_or_insert_with(Default::default);
-        flags.push(" -Z link-native-libraries=no");
-        let emscripten_args = [
-            "-C".to_string(),
-            "link-arg=-sSIDE_MODULE=2".to_string(),
-            "-C".to_string(),
-            "link-arg=-sWASM_BIGINT".to_string(),
-        ];
+        // Allow user to override these default flags
+        if !flags.to_string_lossy().contains("link-native-libraries") {
+            flags.push(" -Z link-native-libraries=no");
+        }
+        let mut emscripten_args = Vec::new();
+        // Allow user to override these default settings
+        if !cargo_rustc
+            .args
+            .iter()
+            .any(|arg| arg.contains("SIDE_MODULE"))
+        {
+            emscripten_args.push("-C".to_string());
+            emscripten_args.push("link-arg=-sSIDE_MODULE=2".to_string());
+        }
+        if !cargo_rustc
+            .args
+            .iter()
+            .any(|arg| arg.contains("WASM_BIGINT"))
+        {
+            emscripten_args.push("-C".to_string());
+            emscripten_args.push("link-arg=-sWASM_BIGINT".to_string());
+        }
         cargo_rustc.args.extend(emscripten_args);
     }
 
@@ -227,6 +287,7 @@ fn compile_target(
                 build.target = vec![target_triple.to_string()];
             }
         } else {
+            build.enable_zig_ar = true;
             let zig_triple = if target.is_linux() && !target.is_musl_target() {
                 match context.platform_tag.iter().find(|tag| tag.is_manylinux()) {
                     Some(PlatformTag::Manylinux { x, y }) => {
@@ -245,12 +306,14 @@ fn compile_target(
     if context.zig {
         // Pass zig command to downstream, eg. python3-dll-a
         if let Ok((zig_cmd, zig_args)) = cargo_zigbuild::Zig::find_zig() {
-            let zig_cmd = if zig_args.is_empty() {
-                zig_cmd
+            if zig_args.is_empty() {
+                build_command.env("ZIG_COMMAND", zig_cmd);
             } else {
-                format!("{} {}", zig_cmd, zig_args.join(" "))
+                build_command.env(
+                    "ZIG_COMMAND",
+                    format!("{} {}", zig_cmd.display(), zig_args.join(" ")),
+                );
             };
-            build_command.env("ZIG_COMMAND", zig_cmd);
         }
     }
 
@@ -328,6 +391,7 @@ fn compile_target(
         .context("Failed to run `cargo rustc`")?;
 
     let mut artifacts = HashMap::new();
+    let mut linked_paths = Vec::new();
 
     let stream = cargo_build
         .stdout
@@ -368,7 +432,22 @@ fn compile_target(
                         .into_iter()
                         .zip(artifact.filenames);
                     for (crate_type, filename) in tuples {
-                        artifacts.insert(crate_type, filename.into());
+                        let artifact = BuildArtifact {
+                            path: filename.into(),
+                            linked_paths: Vec::new(),
+                        };
+                        artifacts.insert(crate_type, artifact);
+                    }
+                }
+            }
+            // See https://doc.rust-lang.org/cargo/reference/external-tools.html#build-script-output
+            cargo_metadata::Message::BuildScriptExecuted(msg) => {
+                for path in msg.linked_paths.iter().map(|p| p.as_str()) {
+                    // `linked_paths` may include a "KIND=" prefix in the string where KIND is the library kind
+                    if let Some(index) = path.find('=') {
+                        linked_paths.push(path[index + 1..].to_string());
+                    } else {
+                        linked_paths.push(path.to_string());
                     }
                 }
             }
@@ -379,6 +458,11 @@ fn compile_target(
         }
     }
 
+    // Add linked_paths to build artifacts
+    for artifact in artifacts.values_mut() {
+        artifact.linked_paths = linked_paths.clone();
+    }
+
     let status = cargo_build
         .wait()
         .expect("Failed to wait on cargo child process");
