@@ -84,14 +84,13 @@ fn path_to_content_type(path: &Path) -> String {
 impl Metadata21 {
     /// Merge metadata with pyproject.toml, where pyproject.toml takes precedence
     ///
-    /// manifest_path must be the directory, not the file
-    fn merge_pyproject_toml(&mut self, manifest_path: impl AsRef<Path>) -> Result<()> {
-        let manifest_path = manifest_path.as_ref();
-        if !manifest_path.join("pyproject.toml").is_file() {
-            return Ok(());
-        }
-        let pyproject_toml =
-            PyProjectToml::new(manifest_path).context("pyproject.toml is invalid")?;
+    /// pyproject_dir must be the directory containing pyproject.toml
+    pub fn merge_pyproject_toml(
+        &mut self,
+        pyproject_dir: impl AsRef<Path>,
+        pyproject_toml: &PyProjectToml,
+    ) -> Result<()> {
+        let pyproject_dir = pyproject_dir.as_ref();
         if let Some(project) = &pyproject_toml.project {
             self.name = project.name.clone();
 
@@ -105,7 +104,7 @@ impl Metadata21 {
 
             match &project.readme {
                 Some(pyproject_toml::ReadMe::RelativePath(readme_path)) => {
-                    let readme_path = manifest_path.join(readme_path);
+                    let readme_path = pyproject_dir.join(readme_path);
                     let description = Some(fs::read_to_string(&readme_path).context(format!(
                         "Failed to read readme specified in pyproject.toml, which should be at {}",
                         readme_path.display()
@@ -122,7 +121,7 @@ impl Metadata21 {
                         bail!("file and text fields of 'project.readme' are mutually-exclusive, only one of them should be specified");
                     }
                     if let Some(readme_path) = file {
-                        let readme_path = manifest_path.join(readme_path);
+                        let readme_path = pyproject_dir.join(readme_path);
                         let description = Some(fs::read_to_string(&readme_path).context(format!(
                                 "Failed to read readme specified in pyproject.toml, which should be at {}",
                                 readme_path.display()
@@ -146,7 +145,7 @@ impl Metadata21 {
                     bail!("file and text fields of 'project.license' are mutually-exclusive, only one of them should be specified");
                 }
                 if let Some(license_path) = file {
-                    let license_path = manifest_path.join(license_path);
+                    let license_path = pyproject_dir.join(license_path);
                     self.license_files.push(license_path);
                 }
                 if let Some(license_text) = text {
@@ -159,7 +158,7 @@ impl Metadata21 {
             // We're already emitting the License-Files metadata without issue.
             // license-files.globs = ["LICEN[CS]E*", "COPYING*", "NOTICE*", "AUTHORS*"]
             let license_include_targets = ["LICEN[CS]E*", "COPYING*", "NOTICE*", "AUTHORS*"];
-            let escaped_manifest_string = glob::Pattern::escape(manifest_path.to_str().unwrap());
+            let escaped_manifest_string = glob::Pattern::escape(pyproject_dir.to_str().unwrap());
             let escaped_manifest_path = Path::new(&escaped_manifest_string);
             for pattern in license_include_targets.iter() {
                 for license_path in
@@ -284,22 +283,20 @@ impl Metadata21 {
     pub fn from_cargo_toml(
         cargo_toml: &CargoToml,
         manifest_path: impl AsRef<Path>,
+        cargo_metadata: &cargo_metadata::Metadata,
     ) -> Result<Metadata21> {
-        let authors = cargo_toml
-            .package
-            .authors
-            .as_ref()
-            .map(|authors| authors.join(", "));
+        let package = cargo_metadata
+            .root_package()
+            .context("Expected cargo to return metadata with root_package")?;
+        let authors = package.authors.join(", ");
 
         let classifiers = cargo_toml.classifiers();
 
-        let author_email = authors.as_ref().and_then(|authors| {
-            if authors.contains('@') {
-                Some(authors.clone())
-            } else {
-                None
-            }
-        });
+        let author_email = if authors.contains('@') {
+            Some(authors.clone())
+        } else {
+            None
+        };
 
         let extra_metadata = cargo_toml.remaining_core_metadata();
 
@@ -307,9 +304,9 @@ impl Metadata21 {
         let mut description_content_type: Option<String> = None;
         // See https://packaging.python.org/specifications/core-metadata/#description
         // and https://doc.rust-lang.org/cargo/reference/manifest.html#the-readme-field
-        if cargo_toml.package.readme == Some("false".to_string()) {
+        if package.readme == Some("false".into()) {
             // > You can suppress this behavior by setting this field to false
-        } else if let Some(ref readme) = cargo_toml.package.readme {
+        } else if let Some(ref readme) = package.readme {
             let readme_path = manifest_path.as_ref().join(readme);
             description = Some(fs::read_to_string(&readme_path).context(format!(
                 "Failed to read Readme specified in Cargo.toml, which should be at {}",
@@ -346,32 +343,36 @@ impl Metadata21 {
                     name.clone()
                 }
             })
-            .unwrap_or_else(|| cargo_toml.package.name.clone());
+            .unwrap_or_else(|| package.name.clone());
         let mut project_url = extra_metadata.project_url.unwrap_or_default();
-        if let Some(repository) = cargo_toml.package.repository.as_ref() {
+        if let Some(repository) = package.repository.as_ref() {
             project_url.insert("Source Code".to_string(), repository.clone());
         }
 
-        let mut metadata = Metadata21 {
+        let metadata = Metadata21 {
             metadata_version: "2.1".to_owned(),
 
             // Mapped from cargo metadata
             name,
-            version: cargo_toml.package.version.clone(),
-            summary: cargo_toml.package.description.clone(),
+            version: package.version.to_string(),
+            summary: package.description.clone(),
             description,
             description_content_type,
-            keywords: cargo_toml
-                .package
-                .keywords
-                .clone()
-                .map(|keywords| keywords.join(",")),
-            home_page: cargo_toml.package.homepage.clone(),
+            keywords: if package.keywords.is_empty() {
+                None
+            } else {
+                Some(package.keywords.join(","))
+            },
+            home_page: package.homepage.clone(),
             download_url: None,
             // Cargo.toml has no distinction between author and author email
-            author: authors,
+            author: if package.authors.is_empty() {
+                None
+            } else {
+                Some(authors)
+            },
             author_email,
-            license: cargo_toml.package.license.clone(),
+            license: package.license.clone(),
             license_files: Vec::new(),
 
             // Values provided through `[project.metadata.maturin]`
@@ -395,9 +396,6 @@ impl Metadata21 {
             gui_scripts: HashMap::new(),
             entry_points: HashMap::new(),
         };
-
-        let manifest_path = manifest_path.as_ref();
-        metadata.merge_pyproject_toml(manifest_path)?;
         Ok(metadata)
     }
 
@@ -408,7 +406,7 @@ impl Metadata21 {
         let mut fields = vec![
             ("Metadata-Version", self.metadata_version.clone()),
             ("Name", self.name.clone()),
-            ("Version", self.get_version_escaped()),
+            ("Version", self.get_pep440_version()),
         ];
 
         let mut add_vec = |name, values: &[String]| {
@@ -505,8 +503,21 @@ impl Metadata21 {
     /// Returns the version encoded according to PEP 427, Section "Escaping
     /// and Unicode"
     pub fn get_version_escaped(&self) -> String {
-        let re = Regex::new(r"[^\w\d.]+").unwrap();
-        re.replace_all(&self.version, "_").to_string()
+        self.get_pep440_version().replace('-', "_")
+    }
+
+    /// Returns the version encoded according to PEP 440
+    ///
+    /// See https://github.com/pypa/setuptools/blob/d90cf84e4890036adae403d25c8bb4ee97841bbf/pkg_resources/__init__.py#L1336-L1345
+    pub fn get_pep440_version(&self) -> String {
+        match pep440::Version::parse(&self.version) {
+            Some(ver) => ver.normalize(),
+            None => {
+                let ver = self.version.replace(' ', ".");
+                let re = Regex::new(r"[^A-Za-z0-9.]+").unwrap();
+                re.replace_all(&ver, "-").to_string()
+            }
+        }
     }
 
     /// Returns the name of the .dist-info directory as defined in the wheel specification
@@ -545,27 +556,41 @@ fn fold_header(text: &str) -> String {
 #[cfg(test)]
 mod test {
     use super::*;
+    use cargo_metadata::MetadataCommand;
     use indoc::indoc;
-    use std::io::Write;
+    use pretty_assertions::assert_eq;
 
-    fn assert_metadata_from_cargo_toml(readme: &str, cargo_toml: &str, expected: &str) {
-        let mut readme_md = tempfile::NamedTempFile::new().unwrap();
+    fn assert_metadata_from_cargo_toml(
+        readme: &str,
+        cargo_toml: &str,
+        expected: &str,
+    ) -> Metadata21 {
+        let crate_dir = tempfile::tempdir().unwrap();
+        let crate_path = crate_dir.path();
+        let manifest_path = crate_path.join("Cargo.toml");
+        fs::create_dir(crate_path.join("src")).unwrap();
+        fs::write(crate_path.join("src/lib.rs"), "").unwrap();
+
+        let readme_path = crate_path.join("README.md");
+        fs::write(&readme_path, readme.as_bytes()).unwrap();
 
         let readme_path = if cfg!(windows) {
-            readme_md.path().to_str().unwrap().replace('\\', "/")
+            readme_path.to_str().unwrap().replace('\\', "/")
         } else {
-            readme_md.path().to_str().unwrap().to_string()
+            readme_path.to_str().unwrap().to_string()
         };
 
-        readme_md.write_all(readme.as_bytes()).unwrap();
-
         let toml_with_path = cargo_toml.replace("REPLACE_README_PATH", &readme_path);
+        fs::write(&manifest_path, &toml_with_path).unwrap();
 
         let cargo_toml_struct: CargoToml = toml_edit::easy::from_str(&toml_with_path).unwrap();
+        let cargo_metadata = MetadataCommand::new()
+            .manifest_path(manifest_path)
+            .exec()
+            .unwrap();
 
         let metadata =
-            Metadata21::from_cargo_toml(&cargo_toml_struct, &readme_md.path().parent().unwrap())
-                .unwrap();
+            Metadata21::from_cargo_toml(&cargo_toml_struct, crate_path, &cargo_metadata).unwrap();
 
         let actual = metadata.to_file_contents().unwrap();
 
@@ -583,11 +608,15 @@ mod test {
                 && cargo_toml.contains("version = \"0.1.0\""),
             "cargo_toml name and version string do not match hardcoded values, test will fail",
         );
-        assert_eq!(
-            metadata.get_dist_info_dir(),
-            PathBuf::from("info_project-0.1.0.dist-info"),
-            "Dist info dir differed from expected"
-        );
+
+        if cargo_toml_struct.remaining_core_metadata().name.is_none() {
+            assert_eq!(
+                metadata.get_dist_info_dir(),
+                PathBuf::from("info_project-0.1.0.dist-info"),
+                "Dist info dir differed from expected"
+            );
+        }
+        metadata
     }
 
     #[test]
@@ -638,7 +667,7 @@ mod test {
             Home-Page: https://example.org
             Author: konstin <konstin@mailbox.org>
             Author-email: konstin <konstin@mailbox.org>
-            Description-Content-Type: text/plain; charset=UTF-8
+            Description-Content-Type: text/markdown; charset=UTF-8; variant=GFM
             Project-URL: Bug Tracker, http://bitbucket.org/tarek/distribute/issues/
 
             # Some test package
@@ -711,6 +740,13 @@ mod test {
 
     #[test]
     fn test_metadata_from_cargo_toml_name_override() {
+        let readme = indoc!(
+            r#"
+            Some test package
+            =================
+        "#
+        );
+
         let cargo_toml = indoc!(
             r#"
             [package]
@@ -719,6 +755,7 @@ mod test {
             version = "0.1.0"
             description = "A test project"
             homepage = "https://example.org"
+            readme = "REPLACE_README_PATH"
 
             [lib]
             crate-type = ["cdylib"]
@@ -744,21 +781,14 @@ mod test {
             Home-Page: https://example.org
             Author: konstin <konstin@mailbox.org>
             Author-email: konstin <konstin@mailbox.org>
+            Description-Content-Type: text/x-rst
+
+            Some test package
+            =================
         "#
         );
 
-        let cargo_toml_struct: CargoToml = toml_edit::easy::from_str(cargo_toml).unwrap();
-        let metadata =
-            Metadata21::from_cargo_toml(&cargo_toml_struct, "/not/exist/manifest/path").unwrap();
-        let actual = metadata.to_file_contents().unwrap();
-
-        assert_eq!(
-            actual.trim(),
-            expected.trim(),
-            "Actual metadata differed from expected\nEXPECTED:\n{}\n\nGOT:\n{}",
-            expected,
-            actual
-        );
+        let metadata = assert_metadata_from_cargo_toml(readme, cargo_toml, expected);
 
         assert_eq!(
             metadata.get_dist_info_dir(),
@@ -791,9 +821,19 @@ mod test {
 
     #[test]
     fn test_merge_metadata_from_pyproject_toml() {
-        let cargo_toml_str = fs_err::read_to_string("test-crates/pyo3-pure/Cargo.toml").unwrap();
+        let manifest_dir = PathBuf::from("test-crates").join("pyo3-pure");
+        let cargo_toml_str = fs_err::read_to_string(manifest_dir.join("Cargo.toml")).unwrap();
         let cargo_toml: CargoToml = toml_edit::easy::from_str(&cargo_toml_str).unwrap();
-        let metadata = Metadata21::from_cargo_toml(&cargo_toml, "test-crates/pyo3-pure").unwrap();
+        let cargo_metadata = MetadataCommand::new()
+            .manifest_path(manifest_dir.join("Cargo.toml"))
+            .exec()
+            .unwrap();
+        let mut metadata =
+            Metadata21::from_cargo_toml(&cargo_toml, &manifest_dir, &cargo_metadata).unwrap();
+        let pyproject_toml = PyProjectToml::new(manifest_dir.join("pyproject.toml")).unwrap();
+        metadata
+            .merge_pyproject_toml(&manifest_dir, &pyproject_toml)
+            .unwrap();
         assert_eq!(
             metadata.summary,
             Some("Implements a dummy function in Rust".to_string())
@@ -832,11 +872,19 @@ mod test {
 
     #[test]
     fn test_merge_metadata_from_pyproject_toml_with_customized_python_source_dir() {
-        let cargo_toml_str =
-            fs_err::read_to_string("test-crates/pyo3-mixed-py-subdir/Cargo.toml").unwrap();
+        let manifest_dir = PathBuf::from("test-crates").join("pyo3-mixed-py-subdir");
+        let cargo_toml_str = fs_err::read_to_string(manifest_dir.join("Cargo.toml")).unwrap();
         let cargo_toml: CargoToml = toml_edit::easy::from_str(&cargo_toml_str).unwrap();
-        let metadata =
-            Metadata21::from_cargo_toml(&cargo_toml, "test-crates/pyo3-mixed-py-subdir").unwrap();
+        let cargo_metadata = MetadataCommand::new()
+            .manifest_path(manifest_dir.join("Cargo.toml"))
+            .exec()
+            .unwrap();
+        let mut metadata =
+            Metadata21::from_cargo_toml(&cargo_toml, &manifest_dir, &cargo_metadata).unwrap();
+        let pyproject_toml = PyProjectToml::new(manifest_dir.join("pyproject.toml")).unwrap();
+        metadata
+            .merge_pyproject_toml(&manifest_dir, &pyproject_toml)
+            .unwrap();
         // defined in Cargo.toml
         assert_eq!(
             metadata.summary,
@@ -848,9 +896,15 @@ mod test {
 
     #[test]
     fn test_implicit_readme() {
-        let cargo_toml_str = fs_err::read_to_string("test-crates/pyo3-mixed/Cargo.toml").unwrap();
+        let manifest_dir = PathBuf::from("test-crates").join("pyo3-mixed");
+        let cargo_toml_str = fs_err::read_to_string(manifest_dir.join("Cargo.toml")).unwrap();
         let cargo_toml = toml_edit::easy::from_str(&cargo_toml_str).unwrap();
-        let metadata = Metadata21::from_cargo_toml(&cargo_toml, "test-crates/pyo3-mixed").unwrap();
+        let cargo_metadata = MetadataCommand::new()
+            .manifest_path(manifest_dir.join("Cargo.toml"))
+            .exec()
+            .unwrap();
+        let metadata =
+            Metadata21::from_cargo_toml(&cargo_toml, &manifest_dir, &cargo_metadata).unwrap();
         assert!(metadata.description.unwrap().starts_with("# pyo3-mixed"));
         assert_eq!(
             metadata.description_content_type.unwrap(),
@@ -860,10 +914,19 @@ mod test {
 
     #[test]
     fn test_merge_metadata_from_pyproject_dynamic_license_test() {
-        let manifest_path = PathBuf::from("test-crates").join("license-test");
-        let cargo_toml_str = fs_err::read_to_string(&manifest_path.join("Cargo.toml")).unwrap();
+        let manifest_dir = PathBuf::from("test-crates").join("license-test");
+        let cargo_toml_str = fs_err::read_to_string(&manifest_dir.join("Cargo.toml")).unwrap();
         let cargo_toml: CargoToml = toml_edit::easy::from_str(&cargo_toml_str).unwrap();
-        let metadata = Metadata21::from_cargo_toml(&cargo_toml, &manifest_path).unwrap();
+        let cargo_metadata = MetadataCommand::new()
+            .manifest_path(manifest_dir.join("Cargo.toml"))
+            .exec()
+            .unwrap();
+        let mut metadata =
+            Metadata21::from_cargo_toml(&cargo_toml, &manifest_dir, &cargo_metadata).unwrap();
+        let pyproject_toml = PyProjectToml::new(manifest_dir.join("pyproject.toml")).unwrap();
+        metadata
+            .merge_pyproject_toml(&manifest_dir, &pyproject_toml)
+            .unwrap();
 
         // verify Cargo.toml value came through
         assert_eq!(metadata.license.as_ref().unwrap(), "MIT");
@@ -872,11 +935,11 @@ mod test {
         assert_eq!(4, metadata.license_files.len());
 
         // Verify pyproject.toml license = {file = ...} worked
-        assert_eq!(metadata.license_files[0], manifest_path.join("LICENCE.txt"));
+        assert_eq!(metadata.license_files[0], manifest_dir.join("LICENCE.txt"));
 
         // Verify the default licenses were included
-        assert_eq!(metadata.license_files[1], manifest_path.join("LICENSE"));
-        assert_eq!(metadata.license_files[2], manifest_path.join("NOTICE.md"));
-        assert_eq!(metadata.license_files[3], manifest_path.join("AUTHORS.txt"));
+        assert_eq!(metadata.license_files[1], manifest_dir.join("LICENSE"));
+        assert_eq!(metadata.license_files[2], manifest_dir.join("NOTICE.md"));
+        assert_eq!(metadata.license_files[3], manifest_dir.join("AUTHORS.txt"));
     }
 }
